/*  plumb - the pipe shell - bytecode executer for plumb scripts
    Copyright (C) 2012-2013  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page: http://repo.hu/projects/plumb
    Contact: plumb (at) igor2.repo.hu
*/

#include <assert.h>
#include "gram_execute.h"
#include "pipe.h"
#include "plumb.h"
#include "var.h"

static void exe_execute(transaction_t *tx, int destroy);

static void tx_flush(transaction_t *tx, int start)
{
	exe_execute(tx, start);
	var_flush(tx, 1);
	var_uninit(tx);
}

void exe_free_fd(exe_fd_t *f)
{
	free(f->proc);
	free(f);
}

exe_fd_t *exe_fdref_(char *id, int fd)
{
	exe_fd_t *ret;
	ret = malloc(sizeof(exe_fd_t));
	ret->proc = id;
	ret->fd = fd;
	return ret;
}

exe_fd_t *exe_fdclone(exe_fd_t *f)
{
	return exe_fdref_(strclone(f->proc), f->fd);
}


exe_fd_t *exe_fdref(char *id, char *fd)
{
	int ifd;
	char *end;

	ifd = strtol(fd, &end, 10);
	if (*end != '\0')
		plumb_error("Invalid integer fd: %s", fd);
	return exe_fdref_(id, ifd);
}

static exe_t *append(parser_t *p, exe_op_t op)
{
	exe_t *e;
	e = malloc(sizeof(exe_t));
	e->op = op;
	if (p->tx->tail == NULL)
		p->tx->head = e;
	else
		p->tx->tail->next = e;

	p->tx->tail = e;
	e->next = NULL;

	return e;
}

void exe_free_create_proc(exe_t *e)
{
	if (e->data.create_proc.id != NULL)
		free(e->data.create_proc.id);
	if (e->data.create_proc.cmd != NULL)
		free(e->data.create_proc.cmd);
	if (e->data.create_proc.args != NULL)
		kvlist_free(e->data.create_proc.args);
	if (e->data.create_proc.props != NULL)
		kvlist_free(e->data.create_proc.props);
}

void exe_append_create_proc(parser_t *p, char *id, int virt, char *cmd, kvlist_t *args, kvlist_t *props)
{
	exe_t *e;
	e = append(p, EXE_CREATE_PROC);
	e->data.create_proc.id    = id;
	e->data.create_proc.virt  = virt;
	e->data.create_proc.cmd   = cmd;
	e->data.create_proc.args  = args;
	e->data.create_proc.props = props;
	e->data.create_proc.proc  = NULL;
}

void exe_do_create_proc(transaction_t *tx, exe_t *e)
{
	exe_create_proc_t *cp = &e->data.create_proc;
	cp->proc = proc_create(tx, cp->id, cp->virt, cp->cmd, cp->args, cp->props);
	cp->id = NULL;
	cp->cmd = NULL;
}

void exe_free_reg_pipe(exe_t *e)
{
	exe_free_fd(e->data.reg_pipe.from);
	exe_free_fd(e->data.reg_pipe.to);
}

void exe_append_reg_pipe(parser_t *p, exe_fd_t *from, exe_fd_t *to)
{
	exe_t *e;
	e = append(p, EXE_REG_PIPE);
	e->data.reg_pipe.from = from;
	e->data.reg_pipe.to   = to;
}

void exe_do_reg_pipe(transaction_t *tx, exe_t *e)
{
	exe_reg_pipe_t *rp = &e->data.reg_pipe;

	pipe_register(fd_ref(rp->from->proc, rp->from->fd), fd_ref(rp->to->proc, rp->to->fd));
}

void exe_free_proc_st(exe_t *e)
{
	free(e->data.proc_st.id);
}

void exe_append_proc_st(parser_t *p, char *id, proc_st_change_t newst)
{
	exe_t *e;
	e = append(p, EXE_PROC_ST);
	e->data.proc_st.id    = id;
	e->data.proc_st.newst = newst;
}

void exe_do_proc_st(transaction_t *tx, exe_t *e)
{
	exe_proc_st_t *ps = &e->data.proc_st;
	proc_virt_t *p = proc_lookup_safe(ps->id);

#warning TODO: delete this when transaction validation is introduced (that will check this before execute)
	if (p == NULL) {
		plumb_error("exe_do_proc_st: can not change process status: process %s does not exist\n", ps->id);
		return;
	}

	switch(ps->newst) {
		case PROC_START:
			if (p->h.state == PS_STOPPED)
				proc_start(p);
			break;
		case PROC_STOP:
			if (p->h.state != PS_STOPPED)
				proc_stop(p, 0);
			break;
		case PROC_KILL:
			if (p->h.state != PS_STOPPED)
				proc_stop(p, 1);
			break;
		case PROC_PAUSE:
			if (p->h.state == PS_RUNNING)
				proc_pause(p, 1);
			break;
		case PROC_RESUME:
			if (p->h.state == PS_PAUSED)
				proc_pause(p, 0);
			break;
	}
}

void exe_append_exit(parser_t *p)
{
	append(p, EXE_EXIT);
}

void exe_do_exit(transaction_t *tx, exe_t *e)
{
	plumb_exit(0);
}

void exe_append_dump_internals(parser_t *p, char *name)
{
	exe_t *e;
	e = append(p, EXE_DUMP_INTERNALS);
	e->data.dump_internals_name = name;
}

void exe_do_dump_internals(transaction_t *tx, exe_t *e)
{
	char *name = e->data.dump_internals_name;
	print_internals(stderr, name == NULL ? "KW_DUMP_INTERNALS" : name);
}

void exe_free_dump_internals(exe_t *e)
{
	if (e->data.dump_internals_name != NULL)
		free(e->data.dump_internals_name);
}

void exe_append_subtx(parser_t *p, transaction_t *tx)
{
	exe_t *e;
	e = append(p, EXE_SUBTX);
	e->data.subtx = tx;
}

void exe_do_subtx(transaction_t *tx, exe_t *e)
{
	tx_flush(e->data.subtx, 1);
}


void exe_append_query(parser_t *p, char *q, char *obj, char *cookie)
{
	exe_t *e;
	e = append(p, EXE_QUERY);
	e->data.query.q      = q;
	e->data.query.obj    = obj;
	e->data.query.cookie = cookie;
}

void exe_do_query(transaction_t *tx, exe_t *e)
{
	exe_query_t *q = &e->data.query;

	if (tx->query_out != NULL)
		run_query(tx->query_out, q->q, q->obj, q->cookie);
}

void exe_free_query(exe_t *e)
{
	free(e->data.query.q);
	free(e->data.query.obj);
	if (e->data.query.cookie != NULL)
		free(e->data.query.cookie);
}

/* === infratructure === */

typedef void (*exe_free_any_t)(exe_t *e);
typedef void (*exe_do_any_t)(transaction_t *tx, exe_t *e);
typedef struct any_s {
	exe_free_any_t  free;
	exe_do_any_t    doit;
} any_t;

static any_t any[] = { /* order does matter: must match exe_op_t */
	{exe_free_dump_internals,exe_do_dump_internals},
	{NULL,                   exe_do_subtx},
	{exe_free_query,         exe_do_query},
	{NULL,                   exe_do_exit},
	{exe_free_proc_st,       exe_do_proc_st},
	{exe_free_reg_pipe,      exe_do_reg_pipe},
	{exe_free_create_proc,   exe_do_create_proc}
};


#define validate_op(e) \
	assert(e->op >= 0); \
	assert(e->op < (sizeof(any) / sizeof(any_t)));

void exe_free_node(exe_t *e)
{
	validate_op(e);
	if (any[e->op].free != NULL)
		any[e->op].free(e);
	free(e);
}

void exe_do_node(transaction_t *tx, exe_t *e)
{
	validate_op(e);
	if (any[e->op].doit != NULL)
		any[e->op].doit(tx, e);
}

/* call do callback of each exe in the chain */
static void exe_do_all(transaction_t *tx)
{
	exe_t *e, *next;
	for(e = tx->head; e != NULL; e = next) {
		next = e->next;
		exe_do_node(tx, e);
	}
}

/* start all processes and free chain if destroy */
static void exe_start_all(transaction_t *tx, int destroy)
{
	exe_t *e, *next;
	for(e = tx->head; e != NULL; e = next) {
		next = e->next;
		if (e->op == EXE_CREATE_PROC)
			proc_start(e->data.create_proc.proc);
		if (destroy)
			exe_free_node(e);
	}
}

static void exe_execute(transaction_t *tx, int start)
{
	exe_do_all(tx);
	if (start) {
		exe_start_all(tx, 1);
		tx->head = tx->tail = NULL;
	}
}

transaction_t *tx_this_or_new(transaction_t *tx, const char *name, fd_t *query_out)
{
	if (tx == NULL) {
		tx = calloc(1, sizeof(transaction_t));
		tx->name = strclone(name);
		tx->query_out = query_out;
	}
	tx->refcnt++;
	return tx;
}

static void tx_free(transaction_t *tx)
{
	if (tx->name != NULL)
		free(tx->name);
	if (tx->err_on_flush != NULL)
		free(tx->err_on_flush);
	free(tx);
}

void exe_flush(parser_t *p)
{
	if (p->tx->err_on_flush != NULL)
		plumb_error("can not flush transaction in exe_flush(): %s\n", p->tx->err_on_flush);
	tx_flush(p->tx, 1);
}

void tx_finish_create(transaction_t *tx)
{
	assert(tx->refcnt == 1);
	tx_flush(tx, 0);
}

void tx_finish_run(transaction_t *tx)
{
	assert(tx->refcnt == 1);
	exe_start_all(tx, 1);
	var_uninit(tx);
	tx_free(tx);
}

transaction_t *tx_finish(transaction_t *tx)
{
	tx->refcnt--;
	if (tx->refcnt == 0) {
		if (tx->err_on_flush != NULL)
			plumb_error("can not flush transaction in tx_finish(): %s\n", tx->err_on_flush);
		tx_flush(tx, 1);
		tx_free(tx);
		return NULL;
	}
	return tx;
}

void tx_discard(transaction_t *tx)
{
	exe_t *e, *next;
	tx->refcnt--;
	for(e = tx->head; e != NULL; e = next) {
		next = e->next;
		exe_free_node(e);
	}
	tx->head = tx->tail = NULL;
	var_uninit(tx);
	if (tx->refcnt == 0)
		tx_free(tx);
}
