/*  plumb - the pipe shell - grammar
    Copyright (C) 2012-2013  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page: http://repo.hu/projects/plumb
    Contact: plumb (at) igor2.repo.hu
*/

%pure-parser
%lex-param {void *scanner, parser_t *tx}
%parse-param {parser_t *yyparser_struct}
%define api.push_pull "both"
%{
#include <stdlib.h>
#include "gram_execute.h"
#include "parse_common.h"
#include "plumb.h"
#include "var.h"
#define YYLEX_PARAM yyparser_struct->scanner, (yyparser_struct)

%}

%union {
	char *str;
	int num;

	exe_fd_t *fd;       /* fd reference (proc/fd) */
	char *proc;         /* name of the process */
	kv_t *kv;           /* key = value */
	kvlist_t *kvlist;   /* list of key-value pairs */
}


%token NL
%token KW_STOP
%token KW_KILL
%token KW_PAUSE
%token KW_RESUME
%token KW_FLUSH
%token KW_EXIT
%token KW_DUMP_INTERNALS
%token KW_QUERY
%token <str> INT
%token <str> STRING

%type <proc>    proc
%type <proc>    proc_virtual
%type <proc>    proc_real
%type <proc>    proc_ref
%type <fd>      pipeline
%type <fd>      fd
%type <str>     opt_id
%type <str>     string
%type <str>     single_string
%type <str>     opt_string
%type <kvlist>  opt_rp_args
%type <kvlist>  opt_vp_args
%type <kvlist>  opt_procprops
%type <kvlist>  procprops
%type <kv>      key_value

%locations
%error-verbose

%%
program:
	  exprseps_or_empty expr exprseps
	| program expr exprseps
	;

exprseps_or_empty:
	  exprseps
	| /* empty */
	;

/* one or more expression separator */
exprseps:
	  exprsep
	| exprseps exprsep
	;

/* expression separator is a semicolon or a newline */
exprsep:
	  ';'
	| NL
	;

/* expression is a pipeline or variable assignment */
expr:
	  pipeline                                      { exe_free_fd($1); }
	| var_assign
	| KW_STOP string                                { exe_append_proc_st(yyparser_struct, $2, PROC_STOP); }
	| KW_KILL string                                { exe_append_proc_st(yyparser_struct, $2, PROC_KILL); }
	| KW_PAUSE string                               { exe_append_proc_st(yyparser_struct, $2, PROC_PAUSE); }
	| KW_RESUME string                              { exe_append_proc_st(yyparser_struct, $2, PROC_RESUME); }
	| KW_FLUSH                                      { exe_flush(yyparser_struct); };
	| KW_EXIT                                       { exe_append_exit(yyparser_struct); };
	| KW_DUMP_INTERNALS opt_string                  { exe_append_dump_internals(yyparser_struct, $2); }
	| KW_QUERY string string opt_string             { exe_append_query(yyparser_struct, $2, $3, $4); }
	;

var_assign:
	string '=' string                               { char *vname = template_prefix(yyparser_struct->templ, $1); var_set(yyparser_struct->tx, vname, $3); free(vname); free($3); }
	;

/* pipeline has at least one pipe with 2 objects */
pipeline:
	  pipeline pipe proc                            { exe_append_reg_pipe(yyparser_struct,$1, exe_fdref_(strclone($3), 0)); $$ = exe_fdref_($3, 1); }
	| pipeline pipe fd                              { $$ = exe_fdclone($3); exe_append_reg_pipe(yyparser_struct,$1, $3); }
	| proc                                          { $$ = exe_fdref_($1, 1); }
	| fd                                            { $$ = $1; }
	;

pipe:
	'|'
	;

/* object is a newly created process or reference to an existing process */
proc:
	  proc_virtual                                  { $$ = $1; }
	| proc_real                                     { $$ = $1; }
	| proc_ref                                      { $$ = $1; }
	;

proc_ref:
	string                                          { $$ = $1;}
	;

/* virtual process (builtins) */
proc_virtual:
	opt_id '[' string opt_vp_args ']' opt_procprops { $$ = $1; exe_append_create_proc(yyparser_struct, strclone($1), 1, $3, $4, $6); }
	;

/* real process (bgproc) */
proc_real:
	opt_id '{' string opt_rp_args '}' opt_procprops { $$ = $1; exe_append_create_proc(yyparser_struct, strclone($1), 0, $3, $4, $6); }
	;

/* explicit fd (proc/int) */
fd:
	  string ':' INT                                { $$ = exe_fdref(template_prefix(yyparser_struct->templ, $1), $3);  free($3); }
	| string ':' '*'                                { $$ = exe_fdref_(template_prefix(yyparser_struct->templ, $1), -1); }
	;

/* optional name for a process */
opt_id:
	  /* empty */                                   { $$ = template_prefix(yyparser_struct->templ, anon_id()); }
	| string '='                                    { $$ = template_prefix(yyparser_struct->templ, $1); }
	;

/* arguments for a process (list of strings, may be empty) */
opt_rp_args:
	  /* empty */                                   { $$ = NULL; }
	| opt_rp_args string                            { $$ = kvlist_append($1, kv($2, NULL)); }
	;

opt_vp_args:
	  /* empty */                                   { $$ = NULL; }
	| opt_vp_args string                            { $$ = kvlist_append($1, kv($2, NULL)); }
	| opt_vp_args key_value                         { $$ = kvlist_append($1, $2); }
	;

string:
	  single_string                                 { $$ = $1; }
	| single_string '+' string                      { $$ = str_aconcat($1, $3); }
	;

opt_string:
	  /* empty */                                   { $$ = NULL; }
	| string                                        { $$ = $1; }
	;

/* an argument (string or integer) */
single_string:
	  STRING                                        { $$ = $1; }
	| INT                                           { $$ = $1; }
	| KW_STOP                                       { $$ = strclone("stop"); }
	| KW_KILL                                       { $$ = strclone("kill"); }
	| KW_PAUSE                                      { $$ = strclone("pause"); }
	| KW_RESUME                                     { $$ = strclone("resume"); }
	| KW_FLUSH                                      { $$ = strclone("flush"); }
	| KW_EXIT                                       { $$ = strclone("exit"); }
	| KW_QUERY                                      { $$ = strclone("query"); }
	| KW_DUMP_INTERNALS                             { $$ = strclone("dump_internals"); }
	;

/* optional process properties */
opt_procprops:
	  /* empty */                                   { $$ = NULL; }
	| procprops                                     { $$ = $1; }
	;

/* a list of process properties (non-empty) */
procprops:
	  ',' key_value                                 { $$ = kvlist_append(NULL, $2); }
	| ',' key_value procprops                       { $$ = kvlist_append($3, $2); }
	;

/* process property is key=value */
key_value:
	string '=' string                               { $$ = kv($1, $3); }
	;

%%
