/*  plumb - the pipe shell - split stream by a set of characters
    Copyright (C) 2012-2013  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page: http://repo.hu/projects/plumb
    Contact: plumb (at) igor2.repo.hu
*/
#include <stdlib.h>
#include <string.h>
#include "process.h"
#include "fd.h"
#include "libporty/util/buf.h"
#include "virt.h"
#include "config.h"

typedef struct {
	P_buf_t *buf;
	char *err_msg_overrun, *err_msg_partial;
	int maxlen;
	char partial_pass;
	char sep1;
	char event;
} split_t;

static virt_ctrl_t split_input(proc_virt_t *proc, int local_fd, const char *data, int len)
{
	char *rec;
	int rlen;
	split_t *s = proc->data;

	if ((s->maxlen > 0) && ((P_buf_get_current_len(s->buf) + len) > s->maxlen)) {
		if (s->event) {
			char *err;
			err = malloc(strlen(proc->h.id) + 64);
			len = sprintf(err, "plumb_split %s buffer grew too large\n", proc->h.id);
			vp_write(proc, vstdevent, err, len);
			free(err);
		}
		if (s->err_msg_overrun != NULL)
			vp_write(proc, vstdout, s->err_msg_overrun, strlen(s->err_msg_overrun));
		return VC_QUIT;
	}

	P_buf_fillbin(s->buf, data, len);
	while((rec = P_buf_getbin(s->buf, &rlen)) != NULL) {
		vp_write(proc, vstdout, rec, rlen);
		P_buf_free(rec);
	}
	return VC_OK;
}

static void split_free(proc_virt_t *proc)
{
	split_t *s = proc->data;

	P_buf_uninit(s->buf);
	if (s->err_msg_overrun != NULL)
		free(s->err_msg_overrun);
	free(s);
}

/* have to take over sink_eof because of partial line handling */
static virt_ctrl_t split_sink_eof(proc_virt_t *proc, int local_fd)
{
	split_t *s = proc->data;
	char *err;
	int len;

	if (P_buf_get_current_len(s->buf) == 0)
		return VC_QUIT;

	if (s->event) {
		err = malloc(strlen(proc->h.id) + 128);
		len = sprintf(err, "plumb_split %s partial record at eof (%s)\n", proc->h.id, s->partial_pass ? "passed" : "discarded");
		vp_write(proc, vstdevent, err, len);
		free(err);
	}

	if (s->err_msg_partial != NULL)
		vp_write(proc, vstdout, s->err_msg_partial, strlen(s->err_msg_partial));

	if (s->partial_pass) {
		char *rec;
		P_buf_fillbin(s->buf, &s->sep1, 1);
		/* no checks required, we will have exactly one record */
		rec = P_buf_getbin(s->buf, &len);
		vp_write(proc, vstdout, rec, len);
		P_buf_free(rec);
	}

	return VC_QUIT;
}


int vp_split_init(proc_virt_t *p, kvlist_t *args)
{
	split_t *s;
	char *end;
	char *spch = "\r\n", *overrun = NULL, *smaxlen = NULL, *spartial = NULL, *partial_msg = NULL, *sevent = NULL;
	vp_args_t vargs[] = {
		{"sep",         &spch},
		{"overrun_msg", &overrun},
		{"partial_msg", &partial_msg},
		{"maxlen",      &smaxlen},
		{"partial",     &spartial},
		{"event",       &sevent},
		{NULL, NULL}
	};

	vp_args(args, vargs);

	s = calloc(1, sizeof(split_t));
	s->buf   = P_buf_init(spch);
	s->sep1  = *spch;
	s->event = vp_istrue(sevent);
	s->err_msg_overrun  = strclone(overrun);
	s->err_msg_partial  = strclone(partial_msg);

	if (smaxlen != NULL) {
		s->maxlen = strtol(smaxlen, &end, 10);
		if (*end != '\0')
			plumb_error("[split] requires integer value for argument maxlen\n");
	}
	else
		s->maxlen = VP_SPLIT_DEFAULT_MAX_LEN;

	if ((spartial != NULL) && (strcmp(spartial, "pass")) == 0)
		s->partial_pass = 1;

	p->input    = split_input;
	p->sink_eof = split_sink_eof;
	p->free     = split_free;
	p->data     = s;

	vp_stdfds_set(p, VP_REQ_QUIT_ON_UNBIND, VP_REQ_QUIT_ON_UNBIND, VP_REQ_UNBOUND, VP_REQ_UNBOUND, s->event ? VP_REQ_QUIT_ON_UNBIND : VP_REQ_UNBOUND, VP_REQ_UNBOUND);

	return 0;
}
